@@ -1,6 +1,7 @@
 /*
- * Copyright 2004 Apache Software Foundation 
- * 
+ * Copyright (C) 2000, 2001, 2013 Gregory Trubetskoy
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Apache Software Foundation
+ *
  * Licensed under the Apache License, Version 2.0 (the "License"); you
  * may not use this file except in compliance with the License.  You
  * may obtain a copy of the License at
@@ -16,9 +17,8 @@
  * Originally developed by Gregory Trubetskoy.
  *
  *
- * connobject.c 
+ * connobject.c
  *
- * $Id: connobject.c 481717 2006-12-03 04:36:37Z grahamd $
  *
  */
 
@@ -70,13 +70,15 @@ static PyObject * conn_log_error(connobject *self, PyObject *args)
     if (message) {
 
         if (! level)
-            level = APLOG_NOERRNO|APLOG_ERR;
+            level = APLOG_ERR;
 
+        Py_BEGIN_ALLOW_THREADS
 #if AP_MODULE_MAGIC_AT_LEAST(20020903,10)
         ap_log_cerror(APLOG_MARK, level, 0, self->conn, "%s", message);
 #else
         ap_log_error(APLOG_MARK, level, 0, self->conn->base_server, "%s", message);
 #endif
+        Py_END_ALLOW_THREADS
     }
 
     Py_INCREF(Py_None);
@@ -109,37 +111,35 @@ static PyObject * _conn_read(conn_rec *c, ap_input_mode_t mode, long len)
         Py_END_ALLOW_THREADS;
 
         if (rc != APR_SUCCESS) {
-            PyErr_SetObject(PyExc_IOError, 
-                            PyString_FromString("Connection read error"));
+            PyErr_SetString(PyExc_IOError, "Connection read error");
             return NULL;
         }
     }
 
-    /* 
-     * loop through the brigade reading buckets into the string 
+    /*
+     * loop through the brigade reading buckets into the string
      */
 
     b = APR_BRIGADE_FIRST(bb);
 
-    if (APR_BUCKET_IS_EOS(b)) { 
+    if (APR_BUCKET_IS_EOS(b)) {
         apr_bucket_delete(b);
         Py_INCREF(Py_None);
         return Py_None;
     }
 
-    /* PYTHON 2.5: 'PyString_FromStringAndSize' uses Py_ssize_t for input parameters */
-    result = PyString_FromStringAndSize(NULL, bufsize);
+    result = PyBytes_FromStringAndSize(NULL, bufsize);
 
     /* possibly no more memory */
-    if (result == NULL) 
+    if (result == NULL)
         return PyErr_NoMemory();
-    
-    buffer = PyString_AS_STRING((PyStringObject *) result);
+
+    buffer = PyBytes_AS_STRING((PyBytesObject *) result);
 
     bytes_read = 0;
 
     while ((bytes_read < len || len == 0) &&
-           !(b == APR_BRIGADE_SENTINEL(b) ||
+           !(b == APR_BRIGADE_SENTINEL(bb) ||
              APR_BUCKET_IS_EOS(b) || APR_BUCKET_IS_FLUSH(b))) {
 
         const char *data;
@@ -147,8 +147,7 @@ static PyObject * _conn_read(conn_rec *c, ap_input_mode_t mode, long len)
         apr_bucket *old;
 
         if (apr_bucket_read(b, &data, &size, APR_BLOCK_READ) != APR_SUCCESS) {
-            PyErr_SetObject(PyExc_IOError, 
-                            PyString_FromString("Connection read error"));
+            PyErr_SetString(PyExc_IOError, "Connection read error");
             return NULL;
         }
 
@@ -165,9 +164,8 @@ static PyObject * _conn_read(conn_rec *c, ap_input_mode_t mode, long len)
         /* time to grow destination string? */
         if (len == 0 && bytes_read == bufsize) {
 
-            /* PYTHON 2.5: '_PyString_Resize' uses Py_ssize_t for input parameters */ 
-            _PyString_Resize(&result, bufsize + HUGE_STRING_LEN);
-            buffer = PyString_AS_STRING((PyStringObject *) result);
+            _PyBytes_Resize(&result, bufsize + HUGE_STRING_LEN);
+            buffer = PyBytes_AS_STRING((PyBytesObject *) result);
             buffer += bufsize;
             bufsize += HUGE_STRING_LEN;
         }
@@ -184,9 +182,8 @@ static PyObject * _conn_read(conn_rec *c, ap_input_mode_t mode, long len)
     }
 
     /* resize if necessary */
-    if (bytes_read < len || len == 0) 
-        /* PYTHON 2.5: '_PyString_Resize' uses Py_ssize_t for input parameters */ 
-        if(_PyString_Resize(&result, bytes_read))
+    if (bytes_read < len || len == 0)
+        if(_PyBytes_Resize(&result, bytes_read))
             return NULL;
 
     return result;
@@ -202,7 +199,7 @@ static PyObject * conn_read(connobject *self, PyObject *args)
 
     long len = 0;
 
-    if (! PyArg_ParseTuple(args, "|l", &len)) 
+    if (! PyArg_ParseTuple(args, "|l", &len))
         return NULL;
 
     if (len == -1)
@@ -221,7 +218,7 @@ static PyObject * conn_readline(connobject *self, PyObject *args)
 
     long len = 0;
 
-    if (! PyArg_ParseTuple(args, "|l", &len)) 
+    if (! PyArg_ParseTuple(args, "|l", &len))
         return NULL;
 
     return _conn_read(self->conn, AP_MODE_GETLINE, len);
@@ -235,26 +232,16 @@ static PyObject * conn_readline(connobject *self, PyObject *args)
 static PyObject * conn_write(connobject *self, PyObject *args)
 {
     char *buff;
-    int len;
+    Py_ssize_t len;
     apr_bucket_brigade *bb;
     apr_bucket *b;
     PyObject *s;
     conn_rec *c = self->conn;
 
-    if (! PyArg_ParseTuple(args, "O", &s)) 
-        return NULL;
-
-    if (! PyString_Check(s)) {
-        PyErr_SetString(PyExc_TypeError, "Argument to write() must be a string");
-        return NULL;
-    }
-
-    /* PYTHON 2.5: 'PyString_Size' uses Py_ssize_t for return values (may need overflow check) */
-    len = PyString_Size(s);
+    if (! PyArg_ParseTuple(args, "s#", &buff, &len))
+        return NULL;  /* bad args */
 
     if (len) {
-        buff = apr_pmemdup(c->pool, PyString_AS_STRING(s), len);
-
         bb = apr_brigade_create(c->pool, c->bucket_alloc);
 
         b = apr_bucket_pool_create(buff, len, c->pool, c->bucket_alloc);
@@ -281,27 +268,34 @@ static PyMethodDef connobjectmethods[] = {
 
 #define OFF(x) offsetof(conn_rec, x)
 
-static struct memberlist conn_memberlist[] = {
-    {"base_server",        T_OBJECT,    0,                       RO},
+static PyMemberDef conn_memberlist[] = {
+    {"base_server",        T_OBJECT,    0,                       READONLY},
     /* XXX vhost_lookup_data? */
     /* XXX client_socket? */
-    {"local_addr",         T_OBJECT,    0,                       RO},
-    {"remote_addr",        T_OBJECT,    0,                       RO},
-    {"remote_ip",          T_STRING,    OFF(remote_ip),          RO},
-    {"remote_host",        T_STRING,    OFF(remote_host),        RO},
-    {"remote_logname",     T_STRING,    OFF(remote_logname),     RO},
-    {"aborted",            T_INT,       0,                       RO},
-    {"keepalive",          T_INT,       0,                       RO},
-    {"double_reverse",     T_INT,       0,                       RO},
-    {"keepalives",         T_INT,       OFF(keepalives),         RO},
-    {"local_ip",           T_STRING,    OFF(local_ip),          RO},
-    {"local_host",         T_STRING,    OFF(local_host),        RO},
-    {"id",                 T_LONG,      OFF(id),                 RO},
+    {"local_addr",         T_OBJECT,    0,                       READONLY},
+#if AP_MODULE_MAGIC_AT_LEAST(20111130,0)
+    {"client_addr",        T_OBJECT,    0,                       READONLY},
+    {"client_ip",          T_STRING,    OFF(client_ip),          READONLY},
+    {"remote_ip",          T_STRING,    OFF(client_ip),          READONLY}, /* bw compat */
+#else
+    {"remote_addr",        T_OBJECT,    0,                       READONLY},
+    {"remote_ip",          T_STRING,    OFF(remote_ip),          READONLY},
+#endif
+    {"remote_host",        T_STRING,    OFF(remote_host),        READONLY},
+    {"remote_logname",     T_STRING,    OFF(remote_logname),     READONLY},
+    {"aborted",            T_INT,       0,                       READONLY},
+    {"keepalive",          T_INT,       0,                       READONLY},
+    {"double_reverse",     T_INT,       0,                       READONLY},
+    {"keepalives",         T_INT,       OFF(keepalives),         READONLY},
+    {"local_addr",         T_OBJECT,    0,                       READONLY},
+    {"local_ip",           T_STRING,    OFF(local_ip),           READONLY},
+    {"local_host",         T_STRING,    OFF(local_host),         READONLY},
+    {"id",                 T_LONG,      OFF(id),                 READONLY},
     /* XXX conn_config? */
-    {"notes",              T_OBJECT,    0,                       RO},
+    {"notes",              T_OBJECT,    0,                       READONLY},
     /* XXX filters ? */
     /* XXX document remain */
-    /*{"remain",             T_LONG,      OFF(remain),             RO},*/
+    /*{"remain",             T_LONG,      OFF(remain),             READONLY},*/
     {NULL}  /* Sentinel */
 };
 
@@ -312,58 +306,13 @@ static struct memberlist conn_memberlist[] = {
  */
 
 static void conn_dealloc(connobject *self)
-{  
+{
     Py_XDECREF(self->base_server);
     Py_XDECREF(self->notes);
     Py_XDECREF(self->hlo);
     PyObject_Del(self);
 }
 
-/**
- ** makeipaddr
- **
- *  utility func to make an ip address
- */
-
-static PyObject *makeipaddr(struct apr_sockaddr_t *addr)
-{
-    char *str = NULL;
-    apr_status_t rc;
-    PyObject *ret = NULL;
-
-    rc = apr_sockaddr_ip_get( &str, addr );
-    if (rc==APR_SUCCESS) {
-        ret = PyString_FromString( str );
-    }
-    else {
-        PyErr_SetString(PyExc_SystemError,"apr_sockaddr_ip_get failure");
-    }
-    
-    return ret; 
-}
-
-/**
- ** makesockaddr
- **
- *  utility func to make a socket address
- */
-
-static PyObject *makesockaddr(struct apr_sockaddr_t *addr)
-{
-    PyObject *addrobj = makeipaddr(addr);
-    PyObject *ret = NULL;
-
-    /* apr_sockaddr_port_get was deprecated and removed in apr 1.x
-     * Access the port directly instead
-     */
-    if (addrobj) {
-        apr_port_t port;
-        port = addr->port; 
-        ret = Py_BuildValue("Oi", addrobj, port );
-        Py_DECREF(addrobj);
-    }
-    return ret;
-}
 
 /**
  ** conn_getattr
@@ -377,10 +326,13 @@ static PyObject * conn_getattr(connobject *self, char *name)
 
     PyObject *res;
 
-    res = Py_FindMethod(connobjectmethods, (PyObject *)self, name);
-    if (res != NULL)
-        return res;
-    
+    PyMethodDef *ml = connobjectmethods;
+    for (; ml->ml_name != NULL; ml++) {
+        if (name[0] == ml->ml_name[0] &&
+            strcmp(name+1, ml->ml_name+1) == 0)
+            return PyCFunction_New(ml, (PyObject*)self);
+    }
+
     PyErr_Clear();
 
    if (strcmp(name, "base_server") == 0) {
@@ -403,19 +355,30 @@ static PyObject * conn_getattr(connobject *self, char *name)
         }
     }
     else if (strcmp(name, "aborted") == 0) {
-        return PyInt_FromLong(self->conn->aborted);
+        return PyLong_FromLong(self->conn->aborted);
     }
     else if (strcmp(name, "keepalive") == 0) {
-        return PyInt_FromLong(self->conn->keepalive);
+        return PyLong_FromLong(self->conn->keepalive);
     }
     else if (strcmp(name, "double_reverse") == 0) {
-        return PyInt_FromLong(self->conn->double_reverse);
+        return PyLong_FromLong(self->conn->double_reverse);
     }
     else if (strcmp(name, "local_addr") == 0) {
         return makesockaddr(self->conn->local_addr);
     }
+#if AP_MODULE_MAGIC_AT_LEAST(20111130,0)
+    else if (strcmp(name, "client_addr") == 0) {
+        return makesockaddr(self->conn->client_addr);
+    }
+    else if (strcmp(name, "remote_addr") == 0) {
+        ap_log_cerror(APLOG_MARK, APLOG_WARNING, 0, self->conn, "%s",
+                      "mod_python: conn.remote_addr deprecated in 2.4, "
+                      "use req.useragent_addr or conn.client_addr");
+        return makesockaddr(self->conn->client_addr);
+#else
     else if (strcmp(name, "remote_addr") == 0) {
         return makesockaddr(self->conn->remote_addr);
+#endif
     }
     else if (strcmp(name, "notes") == 0) {
         Py_INCREF(self->notes);
@@ -426,11 +389,28 @@ static PyObject * conn_getattr(connobject *self, char *name)
         return (PyObject *)self->hlo;
     }
     else if (strcmp(name, "_conn_rec") == 0) {
+#if PY_MAJOR_VERSION == 2 && PY_MINOR_VERSION < 7
         return PyCObject_FromVoidPtr(self->conn, 0);
+#else
+        return PyCapsule_New((void *)self->conn, NULL, NULL);
+#endif
     }
-    else
-        return PyMember_Get((char *)self->conn, conn_memberlist, name);
+    else {
 
+#if AP_MODULE_MAGIC_AT_LEAST(20111130,0)
+        if (strcmp(name, "remote_ip") == 0) {
+            ap_log_cerror(APLOG_MARK, APLOG_WARNING, 0, self->conn, "%s",
+                          "mod_python: conn.remote_ip deprecated in 2.4, "
+                          "use req.useragent_ip or conn.client_ip");
+        }
+#endif
+        PyMemberDef *md = find_memberdef(conn_memberlist, name);
+        if (!md) {
+            PyErr_SetString(PyExc_AttributeError, name);
+            return NULL;
+        }
+        return PyMember_GetOne((char*)self->conn, md);
+    }
 }
 
 /**
@@ -447,27 +427,39 @@ static int conn_setattr(connobject *self, char* name, PyObject* value)
                         "can't delete connection attributes");
         return -1;
     }
-    else
-        return PyMember_Set((char *)self->conn, conn_memberlist, name, value);
-
+    else if (strcmp(name, "keepalive") == 0) {
+        if (! PyLong_Check(value)) {
+            PyErr_SetString(PyExc_TypeError, "keepalive must be a integer");
+            return -1;
+        }
+        self->conn->keepalive = PyLong_AsLong(value);
+        return 0;
+    }
+    else {
+        PyMemberDef *md = find_memberdef(conn_memberlist, name);
+        if (!md) {
+            PyErr_SetString(PyExc_AttributeError, name);
+            return -1;
+        }
+        return PyMember_SetOne((char*)self->conn, md, value);
+    }
 }
 
 PyTypeObject MpConn_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,
-    "mp_conn",
-    sizeof(connobject),
-    0,
-    (destructor) conn_dealloc,       /*tp_dealloc*/
-    0,                               /*tp_print*/
-    (getattrfunc) conn_getattr,      /*tp_getattr*/
-    (setattrfunc) conn_setattr,      /*tp_setattr*/
-    0,                               /*tp_compare*/
-    0,                               /*tp_repr*/
-    0,                               /*tp_as_number*/
-    0,                               /*tp_as_sequence*/
-    0,                               /*tp_as_mapping*/
-    0,                               /*tp_hash*/
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
+    "mp_conn",                       /* tp_name */
+    sizeof(connobject),              /* tp_basicsize */
+    0,                               /* tp_itemsize */
+    (destructor) conn_dealloc,       /* tp_dealloc*/
+    0,                               /* tp_print*/
+    (getattrfunc) conn_getattr,      /* tp_getattr*/
+    (setattrfunc) conn_setattr,      /* tp_setattr*/
+    0,                               /* tp_compare*/
+    0,                               /* tp_repr*/
+    0,                               /* tp_as_number*/
+    0,                               /* tp_as_sequence*/
+    0,                               /* tp_as_mapping*/
+    0,                               /* tp_hash*/
 };
 
 
