@@ -1,6 +1,7 @@
 /*
- * Copyright 2004 Apache Software Foundation 
- * 
+ * Copyright (C) 2000, 2001, 2013 Gregory Trubetskoy
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Apache Software Foundation
+ *
  * Licensed under the Apache License, Version 2.0 (the "License"); you
  * may not use this file except in compliance with the License.  You
  * may obtain a copy of the License at
@@ -13,14 +14,21 @@
  * implied.  See the License for the specific language governing
  * permissions and limitations under the License.
  *
- * $Id: _pspmodule.h 106619 2004-11-25 22:10:52Z nd $
  *
  */
 
 #ifndef __PSP_MODULE_H
 #define __PSP_MODULE_H
 
-void init_psp(void);
+#include "Python.h"
+
+#if PY_MAJOR_VERSION < 3
+#define PyBytes_AsString PyString_AsString
+#define PyBytes_FromString PyString_FromString
+#define MpBytesOrUnicode_FromString PyString_FromString
+#else
+#define MpBytesOrUnicode_FromString PyUnicode_FromString
+#endif
 
 #endif /* __PSP_MODULE_H */
 
