@@ -5,8 +5,9 @@ extern "C" {
 #endif
 
 /*
- * Copyright 2004 Apache Software Foundation 
- * 
+ * Copyright (C) 2000, 2001, 2013 Gregory Trubetskoy
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Apache Software Foundation
+ *
  * Licensed under the Apache License, Version 2.0 (the "License"); you
  * may not use this file except in compliance with the License.  You
  * may obtain a copy of the License at
@@ -24,7 +25,6 @@ extern "C" {
  *
  * tableobject.h
  *
- * $Id: tableobject.h 106619 2004-11-25 22:10:52Z nd $
  *
  */
 
@@ -38,14 +38,21 @@ extern "C" {
         apr_table_t     *table;
         apr_pool_t      *pool;
     } tableobject;
-    
-    extern DL_IMPORT(PyTypeObject) MpTable_Type;
-    extern DL_IMPORT(PyTypeObject) MpTableIter_Type;
-    
-#define MpTable_Check(op) ((op)->ob_type == &MpTable_Type)
-    
-    extern DL_IMPORT(PyObject *) MpTable_FromTable Py_PROTO((apr_table_t *t));
-    extern DL_IMPORT(PyObject *) MpTable_New Py_PROTO((void));
+
+    PyAPI_DATA(PyTypeObject) MpTable_Type;
+    PyAPI_DATA(PyTypeObject) MpTableIter_Type;
+
+#define MpTable_Check(op) (Py_TYPE(op) == &MpTable_Type)
+
+    PyAPI_FUNC(PyObject *) MpTable_FromTable (apr_table_t *t);
+    PyAPI_FUNC(PyObject *) MpTable_New (void);
+
+/* #define DEBUG_TABLES 1 */
+#ifdef DEBUG_TABLES
+#define TABLE_DEBUG(str) printf("mp_table: %s\n", str)
+#else
+#define TABLE_DEBUG(str)
+#endif
 
 #ifdef __cplusplus
 }
