@@ -1,6 +1,7 @@
 /*
- * Copyright 2004 Apache Software Foundation 
- * 
+ * Copyright (C) 2000, 2001, 2013 Gregory Trubetskoy
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Apache Software Foundation
+ *
  * Licensed under the Apache License, Version 2.0 (the "License"); you
  * may not use this file except in compliance with the License.  You
  * may obtain a copy of the License at
@@ -16,15 +17,14 @@
  * Originally developed by Gregory Trubetskoy.
  *
  *
- * tableobject.c 
+ * tableobject.c
  *
- * $Id: tableobject.c 466105 2006-10-20 13:28:02Z jgallacher $
  *
  */
 
 #include "mod_python.h"
 
-/** XXX this is a hack. because apr_table_t 
+/** XXX this is a hack. because apr_table_t
     is not available in a header file */
 #define TABLE_HASH_SIZE 32
 struct apr_table_t {
@@ -49,6 +49,8 @@ PyObject * MpTable_FromTable(apr_table_t *t)
 {
     tableobject *result;
 
+    TABLE_DEBUG("MpTable_FromTable");
+
     result = PyObject_New(tableobject, &MpTable_Type);
     if (! result)
         return PyErr_NoMemory();
@@ -59,7 +61,7 @@ PyObject * MpTable_FromTable(apr_table_t *t)
     return (PyObject *)result;
 }
 
-/** 
+/**
  ** MpTable_New
  **
  *  This returns a new object of built-in type table.
@@ -78,9 +80,11 @@ PyObject * MpTable_New()
     tableobject *t;
     apr_pool_t *p;
 
+    TABLE_DEBUG("MpTable_New");
+
     /* XXX need second arg abort function to report mem error */
     apr_pool_create_ex(&p, NULL, NULL, NULL);
-    
+
     /* two is a wild guess */
     t = (tableobject *)MpTable_FromTable(apr_table_make(p, 2));
 
@@ -97,16 +101,19 @@ PyObject * MpTable_New()
  *      Frees table's memory
  */
 
-static void table_dealloc(register tableobject *self)
-{  
+static void table_dealloc(register void *o)
+{
+    tableobject *self = (tableobject *)o;
+
+    TABLE_DEBUG("table_dealloc");
 
     if (MpTable_Check(self)) {
-        if (self->pool) 
+        if (self->pool)
             apr_pool_destroy(self->pool);
         PyObject_Del(self);
     }
     else
-        self->ob_type->tp_free((PyObject *)self);
+        Py_TYPE(self)->tp_free((PyObject *)self);
 
 }
 
@@ -114,6 +121,7 @@ static void table_dealloc(register tableobject *self)
  ** table_print
  **
  *      prints table like a dictionary
+ *      (Useful when debugging)
  */
 
 static int table_print(register tableobject *self, register FILE *fp, register int flags)
@@ -122,6 +130,8 @@ static int table_print(register tableobject *self, register FILE *fp, register i
     apr_table_entry_t *elts;
     register int i;
 
+    TABLE_DEBUG("table_print");
+
     fprintf(fp, "{");
 
     ah = apr_table_elts(self->table);
@@ -150,7 +160,7 @@ static int table_print(register tableobject *self, register FILE *fp, register i
 /**
  ** table_repr
  **
- *      prints table like a dictionary
+ *      repr table like a dictionary
  */
 
 static PyObject * table_repr(tableobject *self)
@@ -161,41 +171,51 @@ static PyObject * table_repr(tableobject *self)
     apr_table_entry_t *elts;
     int i;
 
-    s = PyString_FromString("{");
+    TABLE_DEBUG("table_repr");
+
+    s = PyBytes_FromString("{");
 
     ah = apr_table_elts (self->table);
     elts = (apr_table_entry_t *) ah->elts;
 
     i = ah->nelts;
     if (i == 0)
-        PyString_ConcatAndDel(&s, PyString_FromString("}"));
+        PyBytes_ConcatAndDel(&s, PyBytes_FromString("}"));
 
     while (i--)
         if (elts[i].key)
         {
-            t = PyString_FromString(elts[i].key);
-            PyString_ConcatAndDel(&s, PyObject_Repr(t));
+            t = PyBytes_FromString(elts[i].key);
+            PyBytes_ConcatAndDel(&s, MpObject_ReprAsBytes(t));;
             Py_XDECREF(t);
 
-            PyString_ConcatAndDel(&s, PyString_FromString(": "));
+            PyBytes_ConcatAndDel(&s, PyBytes_FromString(": "));
 
             if (elts[i].val) {
-              t = PyString_FromString(elts[i].val);
+              t = PyBytes_FromString(elts[i].val);
             } else {
               t = Py_None;
               Py_INCREF(t);
             }
 
-            PyString_ConcatAndDel(&s, PyObject_Repr(t));
+            PyBytes_ConcatAndDel(&s, MpObject_ReprAsBytes(t));
             Py_XDECREF(t);
 
             if (i > 0)
-                PyString_ConcatAndDel(&s, PyString_FromString(", "));
+                PyBytes_ConcatAndDel(&s, PyBytes_FromString(", "));
             else
-                PyString_ConcatAndDel(&s, PyString_FromString("}"));
+                PyBytes_ConcatAndDel(&s, PyBytes_FromString("}"));
         }
 
+#if PY_MAJOR_VERSION < 3
     return s;
+#else
+    {
+        PyObject *str = PyUnicode_FromString(PyBytes_AS_STRING(s));
+        Py_DECREF(s);
+        return str;
+    }
+#endif
 }
 
 /**
@@ -205,9 +225,11 @@ static PyObject * table_repr(tableobject *self)
  *      when you do len(table) in Python.
  */
 
-static int tablelength(tableobject *self) 
-{ 
-    return apr_table_elts(self->table)->nelts;
+static Py_ssize_t tablelength(PyObject *self)
+{
+    TABLE_DEBUG("tablelength");
+
+    return apr_table_elts(((tableobject *)self)->table)->nelts;
 }
 
 /**
@@ -216,7 +238,7 @@ static int tablelength(tableobject *self)
  *      Gets a dictionary item
  */
 
-static PyObject * table_subscript(tableobject *self, register PyObject *key)
+static PyObject * table_subscript(PyObject *self, register PyObject *key)
 {
     char *k;
     const apr_array_header_t *ah;
@@ -224,24 +246,24 @@ static PyObject * table_subscript(tableobject *self, register PyObject *key)
     register int i;
     PyObject *list;
 
-    if (key && !PyString_CheckExact(key)) {
-        PyErr_SetString(PyExc_TypeError,
-                        "table keys must be strings");
+    TABLE_DEBUG("table_subscript");
+
+    MP_ANYSTR_AS_STR(k, key, 1);
+    if (!k) {
+        Py_DECREF(key);
         return NULL;
     }
-    k = PyString_AsString(key);
 
     /* it's possible that we have duplicate keys, so
        we can't simply use apr_table_get since that just
        returns the first match.
     */
 
-    /* PYTHON 2.5: 'PyList_New' uses Py_ssize_t for input parameters */
     list = PyList_New(0);
-    if (!list) 
+    if (!list)
         return NULL;
 
-    ah = apr_table_elts (self->table);
+    ah = apr_table_elts (((tableobject *)self)->table);
     elts = (apr_table_entry_t *) ah->elts;
 
     i = ah->nelts;
@@ -250,20 +272,20 @@ static PyObject * table_subscript(tableobject *self, register PyObject *key)
         if (elts[i].key) {
             if (apr_strnatcasecmp(elts[i].key, k) == 0) {
                 PyObject *v = NULL;
-                if (elts[i].val != NULL) {
-                    v = PyString_FromString(elts[i].val);
-                } else { 
+                if (elts[i].val != NULL)
+                    v = MpBytesOrUnicode_FromString(elts[i].val);
+                else {
                     v = Py_None;
                     Py_INCREF(v);
                 }
-                /* PYTHON 2.5: 'PyList_Insert' uses Py_ssize_t for input parameters */
                 PyList_Insert(list, 0, v);
                 Py_DECREF(v);
             }
         }
 
+    Py_DECREF(key); /* becasue of MP_ANYSTR_AS_STR */
+
     /* if no match */
-    /* PYTHON 2.5: 'PyList_Size' uses Py_ssize_t for return values (may need overflow check) */
     if (PyList_Size(list) == 0) {
         Py_DECREF(list);
         PyErr_SetObject(PyExc_KeyError, key);
@@ -271,9 +293,7 @@ static PyObject * table_subscript(tableobject *self, register PyObject *key)
     }
 
     /* if we got one match */
-    /* PYTHON 2.5: 'PyList_Size' uses Py_ssize_t for return values (may need overflow check) */
     if (PyList_Size(list) == 1) {
-        /* PYTHON 2.5: 'PyList_GetItem' uses Py_ssize_t for input parameters */
         PyObject *v = PyList_GetItem(list, 0);
         Py_INCREF(v);
         Py_DECREF(list);
@@ -294,41 +314,41 @@ static PyObject * table_subscript(tableobject *self, register PyObject *key)
  *      string passed in.
  */
 
-static int table_ass_subscript(tableobject *self, PyObject *key, 
+static int table_ass_subscript(PyObject *self, PyObject *key,
                                PyObject *val)
-{ 
+{
+    char *k, *v;
 
-    char *k;
+    TABLE_DEBUG("table_ass_subscript");
 
-    if (key && !PyString_CheckExact(key)) {
-        PyErr_SetString(PyExc_TypeError,
-                        "table keys must be strings");
+    MP_ANYSTR_AS_STR(k, key, 1);
+    if (!k) {
+        Py_XDECREF(key); /* MP_ANYSTR_AS_STR */
         return -1;
     }
 
-    k = PyString_AsString(key);
-
-    if (val == NULL) {
-        apr_table_unset(self->table, k);
-    }
+    if (val == NULL)
+        apr_table_unset(((tableobject *)self)->table, k);
     else {
-        if (val && !PyString_CheckExact(val)) {
-            PyErr_SetString(PyExc_TypeError,
-                            "table values must be strings");
+        MP_ANYSTR_AS_STR(v, val, 1);
+        if (!v) {
+            Py_XDECREF(key); /* MP_ANYSTR_AS_STR */
+            Py_XDECREF(val); /* MP_ANYSTR_AS_STR */
             return -1;
         }
-        apr_table_set(self->table, k, PyString_AsString(val));
+        apr_table_set(((tableobject *)self)->table, k, v);
     }
+    Py_XDECREF(key); /* MP_ANYSTR_AS_STR */
+    Py_XDECREF(val); /* MP_ANYSTR_AS_STR */
     return 0;
 }
 
 /* table as mapping */
 
 static PyMappingMethods table_as_mapping = {
-    /* PYTHON 2.5: 'inquiry' should be perhaps replaced with 'lenfunc' */
-    (inquiry)       tablelength,           /*mp_length*/
-    (binaryfunc)    table_subscript,       /*mp_subscript*/
-    (objobjargproc) table_ass_subscript,   /*mp_ass_subscript*/
+    tablelength,           /*mp_length*/
+    table_subscript,       /*mp_subscript*/
+    table_ass_subscript,   /*mp_ass_subscript*/
 };
 
 /**
@@ -340,24 +360,23 @@ static PyMappingMethods table_as_mapping = {
 
 static PyObject * table_keys(register tableobject *self)
 {
-
     PyObject *v;
     const apr_array_header_t *ah;
     apr_table_entry_t *elts;
     int i, j;
 
+    TABLE_DEBUG("table_keys");
+
     ah = apr_table_elts(self->table);
     elts = (apr_table_entry_t *) ah->elts;
 
-    /* PYTHON 2.5: 'PyList_New' uses Py_ssize_t for input parameters */
     v = PyList_New(ah->nelts);
 
     for (i = 0, j = 0; i < ah->nelts; i++)
     {
         if (elts[i].key)
         {
-            PyObject *key = PyString_FromString(elts[i].key);
-            /* PYTHON 2.5: 'PyList_SetItem' uses Py_ssize_t for input parameters */
+            PyObject *key = MpBytesOrUnicode_FromString(elts[i].key);
             PyList_SetItem(v, j, key);
             j++;
         }
@@ -380,10 +399,10 @@ static PyObject * table_values(register tableobject *self)
     apr_table_entry_t *elts;
     int i, j;
 
+    TABLE_DEBUG("table_values");
+
     ah = apr_table_elts(self->table);
     elts = (apr_table_entry_t *) ah->elts;
-
-    /* PYTHON 2.5: 'PyList_New' uses Py_ssize_t for input parameters */
     v = PyList_New(ah->nelts);
 
     for (i = 0, j = 0; i < ah->nelts; i++)
@@ -391,13 +410,12 @@ static PyObject * table_values(register tableobject *self)
         if (elts[i].key)
         {
             PyObject *val = NULL;
-            if (elts[i].val != NULL) {
-                val = PyString_FromString(elts[i].val);
-            } else { 
+            if (elts[i].val != NULL)
+                val = MpBytesOrUnicode_FromString(elts[i].val);
+            else {
                 val = Py_None;
                 Py_INCREF(val);
             }
-            /* PYTHON 2.5: 'PyList_SetItem' uses Py_ssize_t for input parameters */
             PyList_SetItem(v, j, val);
             j++;
         }
@@ -420,10 +438,10 @@ static PyObject * table_items(register tableobject *self)
     apr_table_entry_t *elts;
     int i, j;
 
+    TABLE_DEBUG("table_items");
+
     ah = apr_table_elts(self->table);
     elts = (apr_table_entry_t *) ah->elts;
-
-    /* PYTHON 2.5: 'PyList_New' uses Py_ssize_t for input parameters */
     v = PyList_New(ah->nelts);
 
     for (i = 0, j = 0; i < ah->nelts; i++)
@@ -431,7 +449,6 @@ static PyObject * table_items(register tableobject *self)
         if (elts[i].key)
         {
             PyObject *keyval = Py_BuildValue("(s,s)", elts[i].key, elts[i].val);
-            /* PYTHON 2.5: 'PyList_SetItem' uses Py_ssize_t for input parameters */
             PyList_SetItem(v, j, keyval);
             j++;
         }
@@ -454,23 +471,38 @@ static int table_merge(tableobject *a, PyObject *b, int override)
     PyObject *key, *value, *skey, *svalue;
     int status;
 
-    if (keys == NULL)
+    TABLE_DEBUG("table_merge");
+
+    if (keys == NULL) {
+        TABLE_DEBUG("  table_merge: keys NULL");
         return -1;
+    }
 
     iter = PyObject_GetIter(keys);
     Py_DECREF(keys);
-    if (iter == NULL)
+    if (iter == NULL) {
+        TABLE_DEBUG("  table_merge: iter NULL");
         return -1;
-    
+    }
+
     for (key = PyIter_Next(iter); key; key = PyIter_Next(iter)) {
+        char *c_skey;
 
         skey = PyObject_Str(key);
         if (skey == NULL) {
             Py_DECREF(iter);
             Py_DECREF(key);
+            TABLE_DEBUG("  table_merge: skey NULL");
             return -1;
         }
-        if (!override && apr_table_get(a->table, PyString_AsString(skey)) != NULL) {
+        MP_ANYSTR_AS_STR(c_skey, skey, 0);
+        if (!c_skey) {
+            Py_DECREF(key);
+            Py_DECREF(skey);
+            TABLE_DEBUG("  table_merge: c_skey NULL");
+            return -1;
+        }
+        if (!override && apr_table_get(a->table, c_skey) != NULL) {
             Py_DECREF(key);
             Py_DECREF(skey);
             continue;
@@ -481,6 +513,7 @@ static int table_merge(tableobject *a, PyObject *b, int override)
             Py_DECREF(iter);
             Py_DECREF(key);
             Py_DECREF(skey);
+            TABLE_DEBUG("  table_merge: value NULL");
             return -1;
         }
         svalue = PyObject_Str(value);
@@ -489,23 +522,27 @@ static int table_merge(tableobject *a, PyObject *b, int override)
             Py_DECREF(key);
             Py_DECREF(skey);
             Py_DECREF(value);
+            TABLE_DEBUG("  table_merge: svalue NULL");
             return -1;
         }
-        status = table_ass_subscript(a, skey, svalue);
+        status = table_ass_subscript((PyObject *)a, skey, svalue);
         Py_DECREF(key);
         Py_DECREF(value);
         Py_DECREF(skey);
         Py_DECREF(svalue);
         if (status < 0) {
             Py_DECREF(iter);
+            TABLE_DEBUG("  table_merge: status < 0");
             return -1;
         }
     }
     Py_DECREF(iter);
-    if (PyErr_Occurred())
+    if (PyErr_Occurred()) {
         /* Iterator completed, via error */
+        TABLE_DEBUG("  table_merge: PyErr_Occurred()");
         return -1;
-    
+    }
+
     return 0;
 }
 
@@ -516,6 +553,8 @@ static int table_merge(tableobject *a, PyObject *b, int override)
 
 static PyObject *table_update(tableobject *self, PyObject *other)
 {
+    TABLE_DEBUG("table_update");
+
     if (table_merge(self, other, 1) < 0)
         return NULL;
 
@@ -536,12 +575,15 @@ static int table_mergefromseq2(tableobject *self, PyObject *seq2, int override)
     PyObject *item;     /* seq2[i] */
     PyObject *fast;     /* item as a 2-tuple or 2-list */
 
+    TABLE_DEBUG("table_mergefromseq2");
+
     it = PyObject_GetIter(seq2);
     if (it == NULL)
         return -1;
 
     for (i = 0; ; ++i) {
         PyObject *key, *value, *skey, *svalue;
+        char *c_skey;
         int n;
 
         fast = NULL;
@@ -583,9 +625,15 @@ static int table_mergefromseq2(tableobject *self, PyObject *seq2, int override)
             goto Fail;
         }
 
-        if (override || apr_table_get(self->table, 
-                                      PyString_AsString(skey)) == NULL) {
-            int status = table_ass_subscript(self, skey, svalue);
+        MP_ANYSTR_AS_STR(c_skey, skey, 0);
+        if (!c_skey) {
+            Py_DECREF(skey);
+            Py_DECREF(svalue);
+            goto Fail;
+        }
+
+        if (override || apr_table_get(self->table, c_skey) == NULL) {
+            int status = table_ass_subscript((PyObject *)self, skey, svalue);
             if (status < 0) {
                 Py_DECREF(skey);
                 Py_DECREF(svalue);
@@ -618,12 +666,16 @@ static int table_mergefromseq2(tableobject *self, PyObject *seq2, int override)
 static PyObject *table_copy(register tableobject *from)
 {
     tableobject *to = (tableobject *)MpTable_New();
+
+    TABLE_DEBUG("table_copy");
+
     if (to != NULL)
         apr_table_overlap(to->table, from->table, 0);
 
     return (PyObject*)to;
 }
 
+#if PY_MAJOR_VERSION < 3
 /**
  ** table_compare
  **
@@ -631,7 +683,7 @@ static PyObject *table_copy(register tableobject *from)
 
 static int table_compare(tableobject *a, tableobject *b)
 {
-    /* 
+    /*
        we're so lazy that we just copy tables to dicts
        and rely on dict's compare ability. this is not
        the best way to do this to say the least
@@ -640,6 +692,8 @@ static int table_compare(tableobject *a, tableobject *b)
     PyObject *ad, *bd;
     int result;
 
+    TABLE_DEBUG("table_compare");
+
     ad = PyDict_New();
     bd = PyDict_New();
 
@@ -653,16 +707,33 @@ static int table_compare(tableobject *a, tableobject *b)
 
     return result;
 }
+#endif
 
 /**
  ** table_richcompare
  **
  */
 
-static PyObject *table_richcompare(PyObject *v, PyObject *w, int op)
+static PyObject *table_richcompare(PyObject *a, PyObject *b, int op)
 {
-    Py_INCREF(Py_NotImplemented); /* XXX */
-    return Py_NotImplemented;
+
+    PyObject *ad, *bd, *result;
+
+    TABLE_DEBUG("table_richcompare");
+
+    ad = PyDict_New();
+    bd = PyDict_New();
+
+    PyDict_Merge(ad, (PyObject*)a, 0);
+    PyDict_Merge(bd, (PyObject*)b, 0);
+
+    result = PyObject_RichCompare(ad, bd, op);
+
+    Py_DECREF(ad);
+    Py_DECREF(bd);
+
+    return result;
+
 }
 
 /**
@@ -673,26 +744,27 @@ static PyObject *table_richcompare(PyObject *v, PyObject *w, int op)
 static PyObject * table_has_key(tableobject *self, PyObject *key)
 {
 
-    const char *val;
+    const char *k;
 
-    if (!PyString_CheckExact(key)) {
-        PyErr_SetString(PyExc_TypeError,
-                        "table keys must be strings");
+    TABLE_DEBUG("table_has_key");
+
+    MP_ANYSTR_AS_STR(k, key, 1);
+    if (!k) {
+        Py_DECREF(key); /* MP_ANYSTR_AS_STR */
         return NULL;
     }
 
-    val = apr_table_get(self->table, PyString_AsString(key));
-
-    if (val)
-        return PyInt_FromLong(1);
+    if (apr_table_get(self->table, k))
+        return PyLong_FromLong(1);
     else
-        return PyInt_FromLong(0);
+        return PyLong_FromLong(0);
 }
 
 /**
  ** table_get
  **
  *    implements get([failobj]) method
+ *    (only returns the first match)
  */
 
 static PyObject *table_get(register tableobject *self, PyObject *args)
@@ -700,20 +772,28 @@ static PyObject *table_get(register tableobject *self, PyObject *args)
     PyObject *key;
     PyObject *failobj = Py_None;
     PyObject *val = NULL;
-    const char *sval;
+    const char *k, *v;
 
-    if (!PyArg_ParseTuple(args, "S|O:get", &key, &failobj))
+    TABLE_DEBUG("table_get");
+
+    if (!PyArg_ParseTuple(args, "O|O:get", &key, &failobj))
         return NULL;
 
-    sval = apr_table_get(self->table, PyString_AsString(key));
-    
-    if (sval == NULL) {
+    MP_ANYSTR_AS_STR(k, key, 1);
+    if (!k) {
+        Py_DECREF(key); /* MP_ANYSTR_AS_STR */
+        return NULL;
+    }
+
+    v = apr_table_get(self->table, k);
+    if (!v) {
         val = failobj;
         Py_INCREF(val);
     }
     else
-        val = PyString_FromString(sval);
+        val = MpBytesOrUnicode_FromString(v);
 
+    Py_DECREF(key); /* MP_ANYSTR_AS_STR */
     return val;
 }
 
@@ -725,44 +805,47 @@ static PyObject *table_get(register tableobject *self, PyObject *args)
 
 static PyObject *table_setdefault(register tableobject *self, PyObject *args)
 {
-    PyObject *key;
+    int len;
     PyObject *failobj = NULL;
-    PyObject *val = NULL;
-    char *k = NULL;
+    PyObject *key, *val = NULL;
+    char *k = NULL, *f = NULL;
     const char *v = NULL;
 
+    TABLE_DEBUG("table_setdefault");
+
     if (!PyArg_ParseTuple(args, "O|O:setdefault", &key, &failobj))
         return NULL;
 
-    if (!PyString_CheckExact(key)) {
-        PyErr_SetString(PyExc_TypeError,
-                        "table keys must be strings");
+    MP_ANYSTR_AS_STR(k, key, 1);
+    if (!k) {
+        Py_DECREF(key); /* MP_ANYSTR_AS_STR */
         return NULL;
     }
 
-    if (failobj && !PyString_CheckExact(failobj)) {
-        PyErr_SetString(PyExc_TypeError,
-                        "table values must be strings");
-        return NULL;
+    if (failobj) {
+        MP_ANYSTR_AS_STR(f, failobj, 1);
+        if (!f) {
+            Py_DECREF(failobj); /* MP_ANYSTR_AS_ATR */
+            return NULL;
+        }
     }
 
-    k = PyString_AsString(key);
     v = apr_table_get(self->table, k);
-
-    if (v == NULL) {
-        if (failobj == NULL) {
-            apr_table_set(self->table, k, "");
-            val = PyString_FromString("");
+    if (!v) {
+        if (f) {
+            apr_table_set(self->table, k, f);
+            val = failobj;
+            Py_INCREF(val);
         }
         else {
-            apr_table_set(self->table, k, PyString_AsString(failobj));
-            val = failobj;
-            Py_XINCREF(val);
+            apr_table_set(self->table, k, "");
+            v = "";
         }
     }
-    else
-        val = PyString_FromString(v);
 
+    val = MpBytesOrUnicode_FromString(v);
+
+    Py_XDECREF(failobj); /* MP_ANYSTR_AS_ATR */
     return val;
 }
 
@@ -773,6 +856,8 @@ static PyObject *table_setdefault(register tableobject *self, PyObject *args)
 
 static PyObject *table_clear(register tableobject *self)
 {
+    TABLE_DEBUG("table_clear");
+
     apr_table_clear(self->table);
 
     Py_INCREF(Py_None);
@@ -790,6 +875,8 @@ static PyObject *table_popitem(tableobject *self)
     apr_table_entry_t *elts;
     PyObject *res;
 
+    TABLE_DEBUG("table_popitem");
+
     ah = (apr_array_header_t *) apr_table_elts(self->table);
     elts = (apr_table_entry_t *) ah->elts;
 
@@ -817,6 +904,8 @@ static int table_traverse(tableobject *self, visitproc visit, void *arg)
     apr_table_entry_t *elts;
     register int i;
 
+    TABLE_DEBUG("table_traverse");
+
     ah = apr_table_elts (self->table);
     elts = (apr_table_entry_t *) ah->elts;
 
@@ -827,9 +916,9 @@ static int table_traverse(tableobject *self, visitproc visit, void *arg)
             int err;
 
             PyObject *v = NULL;
-            if (elts[i].val != NULL) {
-                v = PyString_FromString(elts[i].val);
-            } else { 
+            if (elts[i].val != NULL)
+                v = MpBytesOrUnicode_FromString(elts[i].val);
+            else {
                 v = Py_None;
                 Py_INCREF(v);
             }
@@ -850,6 +939,8 @@ static int table_traverse(tableobject *self, visitproc visit, void *arg)
 
 static int table_tp_clear(tableobject *self)
 {
+    TABLE_DEBUG("table_tp_clear");
+
     table_clear(self);
     return 0;
 }
@@ -857,19 +948,32 @@ static int table_tp_clear(tableobject *self)
 /**
  ** mp_table_add
  **
- *     this function is equivalent of ap_table_add - 
- *     it can create duplicate entries. 
+ *     this function is equivalent of ap_table_add -
+ *     it can create duplicate entries.
  */
 
 static PyObject * mp_table_add(tableobject *self, PyObject *args)
 {
+    PyObject *key, *val;
+    const char *k, *v;
 
-    const char *val, *key;
+    TABLE_DEBUG("mp_table_add");
 
-    if (! PyArg_ParseTuple(args, "ss", &key, &val))
+    if (! PyArg_ParseTuple(args, "OO", &key, &val))
         return NULL;
 
-    apr_table_add(self->table, key, val);
+    MP_ANYSTR_AS_STR(k, key, 1);
+    MP_ANYSTR_AS_STR(v, val, 1);
+    if ((!k) || (!v)) {
+        Py_DECREF(key); /* MP_ANYSTR_AS_STR */
+        Py_DECREF(val); /* MP_ANYSTR_AS_STR */
+        return NULL;
+    }
+
+    apr_table_add(self->table, k, v);
+
+    Py_DECREF(key); /* MP_ANYSTR_AS_STR */
+    Py_DECREF(val); /* MP_ANYSTR_AS_STR */
 
     Py_INCREF(Py_None);
     return Py_None;
@@ -877,19 +981,22 @@ static PyObject * mp_table_add(tableobject *self, PyObject *args)
 
 typedef PyObject * (*tableselectfunc)(apr_table_entry_t *);
 
-staticforward PyObject *tableiter_new(tableobject *, tableselectfunc);
+static PyObject *tableiter_new(tableobject *, tableselectfunc);
 
 static PyObject *select_key(apr_table_entry_t *elts)
 {
-    return PyString_FromString(elts->key);
+    return MpBytesOrUnicode_FromString(elts->key);
 }
 
 static PyObject *select_value(apr_table_entry_t *elts)
 {
     PyObject *val = NULL;
-    if (elts->val != NULL) {
-        val = PyString_FromString(elts->val);
-    } else { 
+
+    TABLE_DEBUG("select_value");
+
+    if (elts->val != NULL)
+        val = MpBytesOrUnicode_FromString(elts->val);
+    else {
         val = Py_None;
         Py_INCREF(val);
     }
@@ -899,21 +1006,29 @@ static PyObject *select_value(apr_table_entry_t *elts)
 
 static PyObject *select_item(apr_table_entry_t *elts)
 {
+    TABLE_DEBUG("select_item");
+
     return Py_BuildValue("(s,s)", elts->key, elts->val);
 }
 
 static PyObject *table_iterkeys(tableobject *self)
 {
+    TABLE_DEBUG("table_iterkeys");
+
     return tableiter_new(self, select_key);
 }
 
 static PyObject *table_itervalues(tableobject *self)
 {
+    TABLE_DEBUG("table_itervalues");
+
     return tableiter_new(self, select_value);
 }
 
 static PyObject *table_iteritems(tableobject *self)
 {
+    TABLE_DEBUG("table_iteritems");
+
     return tableiter_new(self, select_item);
 }
 
@@ -981,10 +1096,22 @@ static PyMethodDef mp_table_methods[] = {
 
 static int table_contains(tableobject *self, PyObject *key)
 {
-    return apr_table_get(self->table, PyString_AsString(key)) != NULL;
+    char *k;
+    const char *v;
+    int rc;
+
+    TABLE_DEBUG("table_contains");
+
+    MP_ANYSTR_AS_STR(k, key, 1);
+    if (!k) {
+        Py_DECREF(key);
+        return -1;
+    }
+    v = apr_table_get(self->table, k);
+    Py_DECREF(key);
+    return (v != NULL);
 }
 
-/* Hack to implement "key in table" */
 static PySequenceMethods table_as_sequence = {
     0,                                  /* sq_length */
     0,                                  /* sq_concat */
@@ -1005,6 +1132,8 @@ static PySequenceMethods table_as_sequence = {
 
 static PyObject *table_new(PyTypeObject *type, PyObject *args, PyObject *kwds)
 {
+    TABLE_DEBUG("table_new");
+
     return MpTable_New();
 }
 
@@ -1014,6 +1143,8 @@ static int table_init(tableobject *self, PyObject *args, PyObject *kwds)
     static char *kwlist[] = {"items", 0};
     int result = 0;
 
+    TABLE_DEBUG("table_init");
+
     if (!PyArg_ParseTupleAndKeywords(args, kwds, "|O:mp_table",
                                      kwlist, &arg))
         result = -1;
@@ -1029,12 +1160,16 @@ static int table_init(tableobject *self, PyObject *args, PyObject *kwds)
 
 static long table_nohash(PyObject *self)
 {
+    TABLE_DEBUG("table_nohash");
+
     PyErr_SetString(PyExc_TypeError, "mp_table objects are unhashable");
     return -1;
 }
 
 static PyObject *table_iter(tableobject *self)
 {
+    TABLE_DEBUG("table_iter");
+
     return tableiter_new(self, select_key);
 }
 
@@ -1048,16 +1183,19 @@ static char mp_table_doc[] =
 "        d[k] = v";
 
 PyTypeObject MpTable_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,
-    "mp_table",
-    sizeof(tableobject),
-    0,
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
+    "mp_table",                         /* tp_name */
+    sizeof(tableobject),                /* tp_basicsize */
+    0,                                  /* tp_itemsize */
     (destructor)table_dealloc,          /* tp_dealloc */
     (printfunc)table_print,             /* tp_print */
     0,                                  /* tp_getattr */
     0,                                  /* tp_setattr */
+#if PY_MAJOR_VERSION < 3
     (cmpfunc)table_compare,             /* tp_compare */
+#else
+    0,                                  /* tp_reserved */
+#endif
     (reprfunc)table_repr,               /* tp_repr */
     0,                                  /* tp_as_number */
     &table_as_sequence,                 /* tp_as_sequence */
@@ -1089,7 +1227,7 @@ PyTypeObject MpTable_Type = {
     (initproc)table_init,               /* tp_init */
     0,                                  /* tp_alloc */
     table_new,                          /* tp_new */
-    (destructor)table_dealloc,          /* tp_free */
+    table_dealloc,                      /* tp_free */
 };
 
 /* Table iterator type */
@@ -1107,6 +1245,9 @@ typedef struct {
 static PyObject *tableiter_new(tableobject *table, tableselectfunc select)
 {
     tableiterobject *ti;
+
+    TABLE_DEBUG("tableiter_new");
+
     ti = PyObject_NEW(tableiterobject, &MpTableIter_Type);
     if (ti == NULL)
         return NULL;
@@ -1128,13 +1269,15 @@ static PyObject *tableiter_next(tableiterobject *ti, PyObject *args)
 {
     apr_table_entry_t *elts = (apr_table_entry_t *) ti->table->table->a.elts;
 
+    TABLE_DEBUG("tableiter_next");
+
     /* make sure the table hasn't change while being iterated */
 
     if (ti->ti_nelts != ti->table->table->a.nelts) {
         PyErr_SetString(PyExc_RuntimeError,
                         "table changed size during iteration");
         return NULL;
-    } 
+    }
 
     /* return the next key/val */
 
@@ -1150,6 +1293,8 @@ static PyObject *tableiter_next(tableiterobject *ti, PyObject *args)
 
 static PyObject *tableiter_getiter(PyObject *it)
 {
+    TABLE_DEBUG("tableiter_getiter");
+
     Py_INCREF(it);
     return it;
 }
@@ -1164,13 +1309,15 @@ static PyObject *tableiter_iternext(tableiterobject *ti)
 {
     apr_table_entry_t *elts = (apr_table_entry_t *) ti->table->table->a.elts;
 
+    TABLE_DEBUG("tableiter_iternext");
+
     /* make sure the table hasn't change while being iterated */
 
     if (ti->ti_nelts != ti->table->table->a.nelts) {
         PyErr_SetString(PyExc_RuntimeError,
                         "table changed size during iteration");
         return NULL;
-    } 
+    }
 
     /* return the next key/val */
 
@@ -1186,13 +1333,12 @@ static PyObject *tableiter_iternext(tableiterobject *ti)
 }
 
 PyTypeObject MpTableIter_Type = {
-    PyObject_HEAD_INIT(NULL)
-    0,                                  /* ob_size */
-    "dictionary-iterator",                      /* tp_name */
-    sizeof(tableiterobject),                    /* tp_basicsize */
+    PyVarObject_HEAD_INIT(&PyType_Type, 0)
+    "dictionary-iterator",              /* tp_name */
+    sizeof(tableiterobject),            /* tp_basicsize */
     0,                                  /* tp_itemsize */
     /* methods */
-    (destructor)tableiter_dealloc,              /* tp_dealloc */
+    (destructor)tableiter_dealloc,      /* tp_dealloc */
     0,                                  /* tp_print */
     0,                                  /* tp_getattr */
     0,                                  /* tp_setattr */
@@ -1213,7 +1359,7 @@ PyTypeObject MpTableIter_Type = {
     0,                                  /* tp_clear */
     0,                                  /* tp_richcompare */
     0,                                  /* tp_weaklistoffset */
-    (getiterfunc)tableiter_getiter,             /* tp_iter */
+    (getiterfunc)tableiter_getiter,     /* tp_iter */
     (iternextfunc)tableiter_iternext,   /* tp_iternext */
     tableiter_methods,                  /* tp_methods */
     0,                                  /* tp_members */
